/*
 * input plugin for LAMIP software, decode twinvq(version 1) files
 *
 * inspired from the xmms-plugin (by the way a very funny source code)
 * completly re-written from example and now allow to read completly
 * vqf v1.x bad headers (that's not the case with the NTTLib functions)
 *
*/
#include <string.h> // strrchr
#include <strings.h> // strncasecnp
#include <plug_in.h>

#include "twinvq.h"
#include "tvqdec.h"

#define BUFFER_SIZE	1024
#define BITS_INT        (sizeof(int)*8)
#define BYTE_BIT	8
#define CHAR_BITS	8

static int	vqf_init(const lamipPluginHandle*);
static void     vqf_cleanup(void);
static void     vqf_decode(lamipURL*, int);
static void     vqf_set_songinfo(lamipURL*, lamipSonginfo*);


/* VQF utilities take from VQF SDK and adapted to lamip */
static void	frtobuf(float[], short[],int, int);
static int	bread(char*, int, int, lamipURL*);
static int	get_bstm(int*, int, lamipURL*);
static int	get_string(char*, int, lamipURL*);
static int	GetVqInfo(tvqConfInfoSubBlock*, int[], int[], int,
			  INDEX*, lamipURL*);
static int	GetBseInfo(tvqConfInfo*, tvqConfInfoSubBlock*,
			   INDEX*, lamipURL*);
static int	GetGainInfo(tvqConfInfo*, tvqConfInfoSubBlock*,
			    INDEX*, lamipURL*);
static int	GetLspInfo(tvqConfInfo*, INDEX*, lamipURL*);
static int	GetPpcInfo(tvqConfInfo*, INDEX*, lamipURL*);
static int	GetEbcInfo(tvqConfInfo*, tvqConfInfoSubBlock*,
			   INDEX*, lamipURL*);


static int	readable = 0;
static int	ptr = 0;
static int	nbuf = 0;
static char	readbuf[BUFFER_SIZE];
static int*	bits_0[N_INTR_TYPE];
static int*	bits_1[N_INTR_TYPE];


static InputPlugin	vqf_functions;

InputPlugin*	lamip_input_info(void)
{
  vqf_functions.common.name	= "vqf decoder";
  vqf_functions.common.description = "LAMIP vqf-1.x decoder";
  vqf_functions.common.init	= vqf_init;
  vqf_functions.common.cleanup	= vqf_cleanup;
  vqf_functions.decode		= vqf_decode;
  vqf_functions.set_song_info	= vqf_set_songinfo;

  return(&vqf_functions);
}

static int vqf_init(const lamipPluginHandle* plug)
{
  //fprintf(stderr, "VQF: init\n");
  lamip_set_mime_type((lamipPluginHandle*)plug, ".vqf", 0);
  return(1);
}

static void     vqf_cleanup(void)
{
  //fprintf(stderr, "VQF: cleanup\n");
}

static void     vqf_decode(lamipURL* url, int sub_trk)
{
  //lamipURL*	url;
  lamipPCMInfo	pcmi;
  char		buffer[BUFFER_SIZE];	// input buffer
  int		errCode		= 0;
  int		header_size	= 0;
  int		int_buf		= 0;
  int		samplerate	= 0;
  int		frameSize	= 0;
  int		numChannels	= 0;	// frame configuration parameters
  float*	frame		= 0;	// decoder output data array
  short*	buf		= 0;	// file output buffer

  INDEX		vqfIndex;
  headerInfo	vqfHeader;
  tvqConfInfo	vqfConfInfo;
  int		wFlag		= 2;

  /* Init some necessary variables */
  memset(&vqfHeader, 0, sizeof(headerInfo));
  memset(&vqfIndex, 0, sizeof(INDEX));
  memset(&vqfConfInfo, 0, sizeof(tvqConfInfo));
  readable = 0;
  ptr = 0;
  nbuf = 0;
  memset(readbuf, 0, BUFFER_SIZE);
  /*
  url = lamip_url_open(path);
  if(!url){
    fprintf(stderr, "VQF: can't open stream %s\n", path);
    return;
  }
  */
  /* test header */
  memset(buffer, 0, BUFFER_SIZE);
  get_string(buffer, KEYWORD_BYTES+VERSION_BYTES, url);
  if(TvqCheckVersion(buffer) == TVQ_UNKNOWN_VERSION){
    fprintf(stderr, "VQF: Wrong version (maybe v2.x)\n");
    //lamip_url_close(url);
    return;
  }
  fprintf(stderr, "VQF : %s\n", buffer);
  memcpy(vqfHeader.ID, buffer, KEYWORD_BYTES+VERSION_BYTES);
  /* read header size  */
  get_bstm(&header_size, 4*8, url);
  /* read COMM chunk and retreive datas format */
  memset(buffer, 0, BUFFER_SIZE);
  get_string(buffer, 4, url);
  if(strncmp(buffer, "COMM", 4)!=0){
    fprintf(stderr, "VQF: COMM chunk not found\n");
    //lamip_url_close(url);
    return;
  }
  int_buf = 0;
  get_bstm(&int_buf, 4*8, url);// COMM size
  get_bstm(&int_buf, 4*8, url);
  fprintf(stderr, "VQF: channel mode %d\n", int_buf);
  vqfHeader.channelMode = int_buf;
  get_bstm(&int_buf, 4*8, url);
  fprintf(stderr, "VQF: bitrate %d\n", int_buf);
  vqfHeader.bitRate = int_buf;
  get_bstm(&int_buf, 4*8, url);
  fprintf(stderr, "VQF: samplerate %d\n", int_buf);
  vqfHeader.samplingRate = int_buf;
  get_bstm(&int_buf, 4*8, url);
  fprintf(stderr, "VQF: securityLevel %d\n", int_buf);
  vqfHeader.securityLevel = int_buf;

  errCode = TvqInitialize(&vqfHeader, &vqfIndex, 0);
  if(errCode){
    fprintf(stderr, "VQF: Can't initialize the decoder %d\n", errCode);
    //lamip_url_close(url);
    return;
  }
  /* jump to DATA chunck */
  memset(buffer, 0, BUFFER_SIZE);
  get_string(buffer, 4, url); // chunk header
  while(strncmp("DATA", buffer, 4)){
    int_buf = 0;
    get_bstm(&int_buf, 4*8, url); // chunk size
    memset(buffer, 0, BUFFER_SIZE);
    get_string(buffer, int_buf, url); // chunk data
    memset(buffer, 0, BUFFER_SIZE);
    get_string(buffer, 4, url); // chunk header
  }
  /* initialize data buffers */
  TvqGetConfInfo(&vqfConfInfo);
  TvqGetVectorInfo(bits_0, bits_1);
  frameSize = TvqGetFrameSize();
  numChannels = TvqGetNumChannels();
  frame = (float*)malloc(sizeof(float)*(frameSize * numChannels));
  buf   = (short*)malloc(sizeof(short)*(frameSize * numChannels));

  fprintf(stderr, "VQF: numChannel : %d, framesize : %d\n",
	  numChannels, frameSize);

  switch(vqfHeader.samplingRate){
  case 44:
    samplerate = 44100;
    break;
  case 22:
    samplerate = 22050;
    break;
  case 11:
    samplerate = 11025;
    break;
  default:
    samplerate = vqfHeader.samplingRate * 1000;
  }
  pcmi.format = PCM_FORMAT_S16_LE;
  pcmi.samplerate = samplerate;
  pcmi.channels = numChannels;
  lamip_open(&pcmi, 0);
  while(lamip_isContinue()){
    tvqConfInfoSubBlock*	vqfSubBlockInfo = 0;
    int		bitcount = 0;
    int		numFixedBitsPerFrame = TvqGetNumFixedBitsPerFrame();
    int		variableBits = 0;
    int		btype = 0;

    bitcount += get_bstm(&vqfIndex.w_type, vqfConfInfo.BITS_WTYPE, url);
    if(TvqWtypeToBtype(vqfIndex.w_type, &vqfIndex.btype)){
      fprintf(stderr, "VQF: Error unknow Window Type: %d\n", vqfIndex.w_type);
      break;
    }
    btype = vqfIndex.btype;
    vqfSubBlockInfo = &vqfConfInfo.cfg[btype];
    bitcount += variableBits;
    bitcount += GetVqInfo(vqfSubBlockInfo,
			  bits_0[btype], bits_1[btype],
			  variableBits, &vqfIndex, url);
    bitcount += GetBseInfo(&vqfConfInfo, vqfSubBlockInfo, &vqfIndex, url);
    bitcount += GetGainInfo(&vqfConfInfo, vqfSubBlockInfo, &vqfIndex, url);
    bitcount += GetLspInfo(&vqfConfInfo, &vqfIndex, url);
    if(vqfSubBlockInfo->ppc_enable)
      bitcount += GetPpcInfo(&vqfConfInfo, &vqfIndex, url);
    if(vqfSubBlockInfo->ebc_enable)
      bitcount += GetEbcInfo(&vqfConfInfo, vqfSubBlockInfo, &vqfIndex, url);

    if(bitcount!=numFixedBitsPerFrame){
      fprintf(stderr, "VQF: bitcount error %d != %d\n",
	      bitcount, numFixedBitsPerFrame);
      break;
    }
    TvqDecodeFrame(&vqfIndex, frame);
    frtobuf(frame, buf, frameSize, numChannels);
    if(wFlag == 0){
      lamip_writeData(buf, frameSize*numChannels*2);
    }
    wFlag = wFlag-1>0 ? wFlag-1 : 0;
  }
  lamip_close();
  free(frame);
  free(buf);
  TvqTerminate(&vqfIndex);
  //lamip_url_close(url);
  return;
}

static void vqf_set_songinfo(lamipURL* url, lamipSonginfo* songinfo)
{
  //lamipURL*	url;
  char		buffer[BUFFER_SIZE];
  int		int_buf = 0;

  if(url==0 || songinfo==0){
    fprintf(stderr, "VQF: can't handle this file\n");
    return;
  }
  
  readable = 0;
  ptr = 0;
  nbuf = 0;
  memset(readbuf, 0, BUFFER_SIZE);
  /*
  url = lamip_url_open(path);
  if(!url){
    fprintf(stderr, "VQF: set_songinfo can't open url\n");
    return;
  }
  */
  /* test header */
  memset(buffer, 0, BUFFER_SIZE);
  get_string(buffer, KEYWORD_BYTES+VERSION_BYTES, url);
  if(TvqCheckVersion(buffer) == TVQ_UNKNOWN_VERSION){
    fprintf(stderr, "VQF: Wrong version (maybe v2.x)\n");
    //lamip_url_close(url);
    return;
  }
  /* read header size  */
  get_bstm(&int_buf, 4*8, url);
  //fprintf(stderr, "VQF: header_size :  %u \n", header_size);
  /* read COMM chunk and retreive datas format */
  memset(buffer, 0, BUFFER_SIZE);
  get_string(buffer, 4, url);
  while(strncmp("DATA", buffer, 4)){
    int_buf = 0;
    if(!strncmp("COMM", buffer, 4)){
      get_bstm(&int_buf, 4*8, url);// COMM size
      get_bstm(&int_buf, 4*8, url);
      fprintf(stderr, "VQF: channel mode %d\n", int_buf);
      get_bstm(&int_buf, 4*8, url);
      fprintf(stderr, "VQF: bitrate %d\n", int_buf);
      get_bstm(&int_buf, 4*8, url);
      fprintf(stderr, "VQF: samplerate %d\n", int_buf);
      get_bstm(&int_buf, 4*8, url);
      fprintf(stderr, "VQF: securityLevel %d\n", int_buf);
    }else if(!strncmp("NAME", buffer, 4)){
      get_bstm(&int_buf, 4*8, url);// NAME size
      memset(buffer, 0, BUFFER_SIZE);
      get_string(buffer, int_buf, url);
      lamip_songinfo_setTitle(songinfo, buffer);
      fprintf(stderr, "VQF: name=%s\n", buffer);
    }else if(!strncmp("COMT", buffer, 4)){
      get_bstm(&int_buf, 4*8, url);// COMT size
      memset(buffer, 0, BUFFER_SIZE);
      get_string(buffer, int_buf, url);
      fprintf(stderr, "VQF: comment=%s\n", buffer);
    }else if(!strncmp("AUTH", buffer, 4)){
      get_bstm(&int_buf, 4*8, url);// AUTH size
      memset(buffer, 0, BUFFER_SIZE);
      get_string(buffer, int_buf, url);
      lamip_songinfo_setArtist(songinfo, buffer);
      fprintf(stderr, "VQF: author=%s\n", buffer);
    }else if(!strncmp("(c) ", buffer, 4)){
      get_bstm(&int_buf, 4*8, url);// (c) size
      memset(buffer, 0, BUFFER_SIZE);
      get_string(buffer, int_buf, url);
      fprintf(stderr, "VQF: copyright=%s\n", buffer);
    }else if(!strncmp("FILE", buffer, 4)){
      get_bstm(&int_buf, 4*8, url);// FILE size
      memset(buffer, 0, BUFFER_SIZE);
      get_string(buffer, int_buf, url);
      fprintf(stderr, "VQF: file=%s\n", buffer);
    }else if(!strncmp("DSIZ", buffer, 4)){
      get_bstm(&int_buf, 4*8, url);// DSIZ size
      get_bstm(&int_buf, 4*8, url);// DSIZ value
      fprintf(stderr, "VQF: size=%d\n", int_buf);
    }else{
      fprintf(stderr, "VQF: can't handle header correctly");
      /* jump the chunck */
      get_bstm(&int_buf, 4*8, url);
      memset(buffer, 0, BUFFER_SIZE);
      get_string(buffer, int_buf, url);
      break;
    }
    memset(buffer, 0, BUFFER_SIZE);
    get_string(buffer, 4, url);    
  }//while
  //lamip_url_close(url);
  return;
}


/***************************************************************************
 *
 * Taken from the TwinVQ SDK directly
*/

static void frtobuf(float out[], short bufout[],
		    int frameSize, int numChannels)
{
  int ismp, ich, uflag, lflag;
  float *ptr;
  register float dtmp;

  uflag = lflag = 0;
  for(ich=0; ich<numChannels; ich++){
    ptr = out+ich*frameSize;
    for(ismp=0; ismp<frameSize; ismp++){
      dtmp = ptr[ismp];
      if(dtmp >= 0.){
	if(dtmp > 32700.){
	  dtmp = 32700.;
	  uflag = 1;
	}
	bufout[ismp*numChannels+ich] = (short)(dtmp+0.5);
      }else{
	if(dtmp < -32700.){
	  dtmp = -32700.;
	  lflag = 1;
	}
	bufout[ismp*numChannels+ich] = (short)(dtmp-0.5);
      }
    }
  }
}


/*
 * Vqf datas are written in a completly strange bit order
 * so use this function to read datas
*/
static int bread(char*	data,    /* Output: Output data array */
		 int	size,     /* Input:  Length of each data */
		 int	nbits,    /* Input:  Number of bits to write */
		 lamipURL* url) /* Input:  File pointer */
{
  int  ibits, iptr, idata, ibufadr, ibufbit, icl;
  unsigned char mask, tmpdat;
  int  retval;

  retval = 0;
  mask = 0x1;
  for(ibits=0; ibits<nbits; ibits++){
    if(readable == 0){  /* when the file data buffer is empty */
      //fprintf(stderr, "VQF: lamip_url_read %d", BUFFER_SIZE);
      nbuf = lamip_url_read(url, readbuf, BUFFER_SIZE);
      //fread(stream->buf, 1, BBUFSIZ, stream->fp);
      nbuf *= 8;
      readable = 1;
    }
    iptr = ptr;      // current file data buffer pointer
    if(iptr>=nbuf){   // If data file is empty then return
      return(retval);
    }
    ibufadr = iptr/BYTE_BIT;      /* current file data buffer address */
    ibufbit = iptr%BYTE_BIT;      /* current file data buffer bit */
    /*      tmpdat = stream->buf[ibufadr] >> (BYTE_BIT-ibufbit-1); */
    tmpdat = (unsigned char)readbuf[ibufadr];
    tmpdat >>= (BYTE_BIT-ibufbit-1);
    /* current data bit */
    idata = ibits*size;                   /* output data address */
    data[idata] = (char)(tmpdat & mask);  /* set output data */
    for(icl=1; icl<size; icl++)
      data[idata+icl] = 0; /* clear the rest output data buffer */
    ptr += 1;       /* update data buffer pointer */
    if(ptr == (BUFFER_SIZE*8)){
      ptr = 0;
      readable = 0;
    }
    ++retval;
  }
  return(retval);
}


static int get_bstm(int*	data,	/* Input: input data */
		    int		nbits,	/* Input: number of bits */
		    lamipURL*	bfp)	/* Input: bit file pointer */
{
  int         ibit;
  unsigned    mask;
  unsigned    work;
  char        tmpbit[BITS_INT];
  int         retval;
        
  if(nbits > (int)BITS_INT){
    fprintf(stderr, "VQF : get_bstm(): %d: %d Error.\n", nbits, BITS_INT);
    //exit(1);
    return(-1);
  }
  retval = bread(tmpbit, sizeof(*tmpbit), nbits, bfp);
  for(ibit=retval; ibit<nbits; ibit++){
    tmpbit[ibit] = 0;
  }
  mask = 0x1<<(nbits-1);
  work=0;
  for( ibit=0; ibit<nbits; ibit++){
    work += mask*tmpbit[ibit];
    mask >>= 1;
  }
  *data = work;
  return(retval);
}


static int get_string(char* buf, int nbytes, lamipURL* bfp)
{
  int c, ichar, ibit;
        
  for(ichar=0; ichar<nbytes; ichar++){
    ibit = get_bstm(&c, CHAR_BITS, bfp);
    if(ibit < CHAR_BITS){
      break;
    }
    buf[ichar] = c;
  }
  buf[ichar] = '\0';
  return(ichar);
}


/* get VQ information
 * Return:      (int) number of bits read
*/
static int GetVqInfo(tvqConfInfoSubBlock* cfg,
		     int bits0[],
		     int bits1[],
		     int variableBits,
		     INDEX* index,
		     lamipURL* bfp)
{
  int idiv;
  int bitcount = 0;

  if(index->btype == BLK_LONG){
    /* re-calculate VQ bits */
    TvqUpdateVectorInfo(variableBits, &cfg->ndiv, bits0, bits1 );
  }
  for(idiv=0; idiv<cfg->ndiv; idiv++){
    /* CB 0 */
    bitcount += get_bstm(&index->wvq[idiv], bits0[idiv], bfp);
    /* CB 1 */
    bitcount += get_bstm(&index->wvq[idiv + cfg->ndiv], bits1[idiv], bfp );
  }
  return(bitcount);
}



/* Description: get BSE information
 * Return:      (int) number of bits read
*/
static int GetBseInfo(tvqConfInfo* cf,
		      tvqConfInfoSubBlock* cfg,
		      INDEX* index, lamipURL* bfp)
{
  int i_sup, isf, itmp, idiv;
  int bitcount = 0;

  for(i_sup=0; i_sup<cf->N_CH; i_sup++){
    for(isf=0; isf<cfg->nsf; isf++){
      for(idiv=0; idiv<cfg->fw_ndiv; idiv++){
	itmp = idiv + ( isf + i_sup * cfg->nsf ) * cfg->fw_ndiv;
	bitcount += get_bstm( &index->fw[itmp], cfg->fw_nbit, bfp );
      }
    }
  }
  for(i_sup=0; i_sup<cf->N_CH; i_sup++){
    for(isf=0; isf<cfg->nsf; isf++){
      bitcount += get_bstm( &index->fw_alf[i_sup*cfg->nsf+isf], cf->FW_ARSW_BITS, bfp);
    }
  }
  return(bitcount);
}


/* get gain information
 * Return:      (int) number of bits read
 */
static int GetGainInfo(tvqConfInfo* cf,
		       tvqConfInfoSubBlock* cfg,
		       INDEX* index, lamipURL* bfp )
{
  int i_sup, iptop, isf;
  int bitcount = 0;

  for(i_sup=0; i_sup<cf->N_CH; i_sup++){
    iptop = ( cfg->nsubg + 1 ) * i_sup;
    bitcount += get_bstm( &index->pow[iptop], cf->GAIN_BITS, bfp );
    for(isf=0; isf<cfg->nsubg; isf++){
      bitcount += get_bstm( &index->pow[iptop+isf+1], cf->SUB_GAIN_BITS, bfp);
    }
  }
  return(bitcount);
}

/* get LSP information
 * Return:      (int) number of bits read
 */
static int GetLspInfo(tvqConfInfo* cf, INDEX* index, lamipURL* bfp)
{
  int i_sup, itmp;
  int bitcount = 0;

  for(i_sup=0; i_sup<cf->N_CH; i_sup++){
    /* pred. switch */
    bitcount += get_bstm( &index->lsp[i_sup][0], cf->LSP_BIT0, bfp);
    /* first stage */
    bitcount += get_bstm( &index->lsp[i_sup][1], cf->LSP_BIT1, bfp);
    for(itmp=0; itmp<cf->LSP_SPLIT; itmp++){
      /* second stage */
      bitcount += get_bstm( &index->lsp[i_sup][itmp+2], cf->LSP_BIT2, bfp);
    }
  }
  return(bitcount);
}

/* get PPC information
 * Return:      (int) number of bits read
 */
static int GetPpcInfo(tvqConfInfo* cf, INDEX* index, lamipURL* bfp)
{
  int idiv, i_sup;
  int bitcount = 0;
        
  for(idiv=0; idiv<cf->N_DIV_P; idiv++){
    /*CB0*/
    bitcount += get_bstm( &(index->pls[idiv]), bits_0[BLK_PPC][idiv], bfp );
    /*CB1*/
    bitcount += get_bstm( &(index->pls[idiv+cf->N_DIV_P]), bits_1[BLK_PPC][idiv], bfp);
  }
  for(i_sup=0; i_sup<cf->N_CH; i_sup++){
    bitcount += get_bstm(&(index->pit[i_sup]), cf->BASF_BIT, bfp);
    bitcount += get_bstm(&(index->pgain[i_sup]), cf->PGAIN_BIT, bfp);
  }     
  return(bitcount);
}


/* get EBC information
 * Return:      (int) number of bits read
 */
static int GetEbcInfo(tvqConfInfo* cf, tvqConfInfoSubBlock* cfg,
		      INDEX* index, lamipURL* bfp)
{
  int i_sup, isf, itmp;
  int bitcount = 0;

  for(i_sup=0; i_sup<cf->N_CH; i_sup++){
    for(isf=0; isf<cfg->nsf; isf++){
      int indexSfOffset = isf * (cfg->ncrb - cfg->ebc_crb_base) - cfg->ebc_crb_base;
      for(itmp=cfg->ebc_crb_base; itmp<cfg->ncrb; itmp++){
	bitcount += get_bstm( &index->bc[i_sup][itmp+indexSfOffset], cfg->ebc_bits, bfp);
      }
    }
  }
  return(bitcount);
}

